@@ -14,6 +14,22 @@
 
 import itertools
 
+import netaddr
+
+from neutron.db import api as db_api
+from neutron.db import common_db_mixin as common_db
+from neutron.db import l3_dvr_db
+from neutron.db.models import address_scope as address_scope_db
+from neutron.db.models import l3 as l3_db
+from neutron.db.models import l3_attrs as l3_attrs_db
+from neutron.db import models_v2
+from neutron.plugins.ml2 import models as ml2_models
+
+from neutron_lib.api import validators
+from neutron_lib import constants as lib_consts
+from neutron_lib.db import model_base
+from neutron_lib import exceptions as n_exc
+from neutron_lib.exceptions import l3 as l3_exc
 from oslo_db import exception as oslo_db_exc
 from oslo_utils import uuidutils
 import sqlalchemy as sa
@@ -22,19 +38,6 @@ from sqlalchemy import orm
 from sqlalchemy.orm import aliased
 from sqlalchemy.orm import exc as sa_exc
 
-from neutron_lib.api import validators
-from neutron_lib import constants as lib_consts
-from neutron_lib.db import model_base
-from neutron_lib import exceptions as n_exc
-
-from neutron.api.v2 import attributes as attr
-from neutron.db import common_db_mixin as common_db
-from neutron.db import l3_attrs_db
-from neutron.db import l3_db
-from neutron.db.models import address_scope as address_scope_db
-from neutron.db import models_v2
-from neutron.plugins.ml2 import models as ml2_models
-
 from neutron_dynamic_routing._i18n import _
 from neutron_dynamic_routing.extensions import bgp as bgp_ext
 
@@ -89,8 +92,8 @@ class BgpSpeaker(model_base.BASEV2,
 
     __tablename__ = 'bgp_speakers'
 
-    name = sa.Column(sa.String(attr.NAME_MAX_LEN), nullable=False)
-    local_as = sa.Column(sa.Integer, nullable=False, autoincrement=False)
+    name = sa.Column(sa.String(255), nullable=False)
+    local_as = sa.Column(sa.BigInteger(), nullable=False, autoincrement=False)
     advertise_floating_ip_host_routes = sa.Column(sa.Boolean, nullable=False)
     advertise_tenant_networks = sa.Column(sa.Boolean, nullable=False)
     peers = orm.relationship(BgpSpeakerPeerBinding,
@@ -112,10 +115,10 @@ class BgpPeer(model_base.BASEV2,
 
     __tablename__ = 'bgp_peers'
 
-    name = sa.Column(sa.String(attr.NAME_MAX_LEN), nullable=False)
+    name = sa.Column(sa.String(255), nullable=False)
     peer_ip = sa.Column(sa.String(64),
                         nullable=False)
-    remote_as = sa.Column(sa.Integer, nullable=False, autoincrement=False)
+    remote_as = sa.Column(sa.BigInteger(), nullable=False, autoincrement=False)
     auth_type = sa.Column(sa.String(16), nullable=False)
     password = sa.Column(sa.String(255), nullable=True)
 
@@ -130,7 +133,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def get_bgp_speakers(self, context, filters=None, fields=None,
                          sorts=None, limit=None, marker=None,
                          page_reverse=False):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             return self._get_collection(context, BgpSpeaker,
                                         self._make_bgp_speaker_dict,
                                         filters=filters, fields=fields,
@@ -138,15 +141,15 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                         page_reverse=page_reverse)
 
     def get_bgp_speaker(self, context, bgp_speaker_id, fields=None):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             bgp_speaker = self._get_bgp_speaker(context, bgp_speaker_id)
             return self._make_bgp_speaker_dict(bgp_speaker, fields)
 
     def get_bgp_speaker_with_advertised_routes(self, context,
                                                bgp_speaker_id):
         bgp_speaker_attrs = ['id', 'local_as', 'tenant_id']
-        bgp_peer_attrs = ['peer_ip', 'remote_as', 'password']
-        with context.session.begin(subtransactions=True):
+        bgp_peer_attrs = ['peer_ip', 'remote_as', 'auth_type', 'password']
+        with db_api.context_manager.reader.using(context):
             bgp_speaker = self.get_bgp_speaker(context, bgp_speaker_id,
                                                fields=bgp_speaker_attrs)
             res = dict((k, bgp_speaker[k]) for k in bgp_speaker_attrs)
@@ -160,7 +163,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def update_bgp_speaker(self, context, bgp_speaker_id, bgp_speaker):
         bp = bgp_speaker[bgp_ext.BGP_SPEAKER_BODY_KEY_NAME]
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             bgp_speaker_db = self._get_bgp_speaker(context, bgp_speaker_id)
             bgp_speaker_db.update(bp)
 
@@ -170,7 +173,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _save_bgp_speaker(self, context, bgp_speaker, uuid):
         ri = bgp_speaker[bgp_ext.BGP_SPEAKER_BODY_KEY_NAME]
         ri['tenant_id'] = context.tenant_id
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             res_keys = ['local_as', 'tenant_id', 'name', 'ip_version',
                         'advertise_floating_ip_host_routes',
                         'advertise_tenant_networks']
@@ -195,7 +198,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def add_gateway_network(self, context, bgp_speaker_id, network_info):
         network_id = self._get_id_for(network_info, 'network_id')
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             try:
                 self._save_bgp_speaker_network_binding(context,
                                                        bgp_speaker_id,
@@ -207,7 +210,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return {'network_id': network_id}
 
     def remove_gateway_network(self, context, bgp_speaker_id, network_info):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             network_id = self._get_id_for(network_info, 'network_id')
             self._remove_bgp_speaker_network_binding(context,
                                                      bgp_speaker_id,
@@ -215,7 +218,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return {'network_id': network_id}
 
     def delete_bgp_speaker(self, context, bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             bgp_speaker_db = self._get_bgp_speaker(context, bgp_speaker_id)
             context.session.delete(bgp_speaker_db)
 
@@ -226,7 +229,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         if auth_type == 'md5' and not password:
             raise bgp_ext.InvalidBgpPeerMd5Authentication()
 
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             res_keys = ['tenant_id', 'name', 'remote_as', 'peer_ip',
                         'auth_type', 'password']
             res = dict((k, ri[k]) for k in res_keys)
@@ -249,23 +252,23 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                      bgp_speaker_id, fields=None):
         filters = [BgpSpeakerPeerBinding.bgp_speaker_id == bgp_speaker_id,
                    BgpSpeakerPeerBinding.bgp_peer_id == BgpPeer.id]
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(BgpPeer)
             query = query.filter(*filters)
-            return [self._make_bgp_peer_dict(x) for x in query.all()]
+            return [self._make_bgp_peer_dict(x, fields) for x in query.all()]
 
     def get_bgp_peer(self, context, bgp_peer_id, fields=None):
         bgp_peer_db = self._get_bgp_peer(context, bgp_peer_id)
         return self._make_bgp_peer_dict(bgp_peer_db, fields=fields)
 
     def delete_bgp_peer(self, context, bgp_peer_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             bgp_peer_db = self._get_bgp_peer(context, bgp_peer_id)
             context.session.delete(bgp_peer_db)
 
     def update_bgp_peer(self, context, bgp_peer_id, bgp_peer):
         bp = bgp_peer[bgp_ext.BGP_PEER_BODY_KEY_NAME]
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             bgp_peer_db = self._get_bgp_peer(context, bgp_peer_id)
             if ((bp.get('password') is not None) and
                 (bgp_peer_db['auth_type'] == 'none')):
@@ -283,7 +286,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
             raise bgp_ext.BgpSpeakerNotFound(id=bgp_speaker_id)
 
     def _get_bgp_speaker_ids_by_router(self, context, router_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             network_binding = aliased(BgpSpeakerNetworkBinding)
             r_port = aliased(l3_db.RouterPort)
             query = context.session.query(network_binding.bgp_speaker_id)
@@ -296,7 +299,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
             return [binding.bgp_speaker_id for binding in query.all()]
 
     def _get_bgp_speaker_ids_by_binding_network(self, context, network_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(
                                       BgpSpeakerNetworkBinding.bgp_speaker_id)
             query = query.filter(
@@ -319,7 +322,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return uuid
 
     def _get_bgp_peers_by_bgp_speaker_binding(self, context, bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(BgpPeer)
             query = query.filter(
                      BgpSpeakerPeerBinding.bgp_speaker_id == bgp_speaker_id,
@@ -328,7 +331,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _save_bgp_speaker_peer_binding(self, context, bgp_speaker_id,
                                        bgp_peer_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             try:
                 bgp_speaker = self._get_by_id(context, BgpSpeaker,
                                               bgp_speaker_id)
@@ -358,7 +361,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _remove_bgp_speaker_peer_binding(self, context, bgp_speaker_id,
                                          bgp_peer_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
 
             try:
                 binding = self._get_bgp_speaker_peer_binding(context,
@@ -374,7 +377,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                           context,
                                           bgp_speaker_id,
                                           network_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             try:
                 bgp_speaker = self._get_by_id(context, BgpSpeaker,
                                               bgp_speaker_id)
@@ -395,7 +398,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _remove_bgp_speaker_network_binding(self, context,
                                             bgp_speaker_id, network_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
 
             try:
                 binding = self._get_bgp_speaker_network_binding(
@@ -449,7 +452,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return self._fields(res, fields)
 
     def _get_address_scope_ids_for_bgp_speaker(self, context, bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             binding = aliased(BgpSpeakerNetworkBinding)
             address_scope = aliased(address_scope_db.AddressScope)
             query = context.session.query(address_scope)
@@ -463,7 +466,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def get_routes_by_bgp_speaker_id(self, context, bgp_speaker_id):
         """Get all routes that should be advertised by a BgpSpeaker."""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             net_routes = self._get_tenant_network_routes_by_bgp_speaker(
                                                                context,
                                                                bgp_speaker_id)
@@ -478,7 +481,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def get_routes_by_bgp_speaker_binding(self, context,
                                           bgp_speaker_id, network_id):
         """Get all routes for the given bgp_speaker binding."""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             fip_routes = self._get_central_fip_host_routes_by_binding(
                                                                context,
                                                                network_id,
@@ -517,7 +520,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_router(self, context, router_id,
                                                bgp_speaker_id):
         """Get floating IP host routes with the given router as nexthop."""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             dest_alias = aliased(l3_db.FloatingIP,
                                  name='destination')
             next_hop_alias = aliased(models_v2.IPAllocation,
@@ -552,7 +555,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_dvr_fip_host_routes_by_router(self, context, bgp_speaker_id,
                                            router_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
 
             fip_query = self._get_fip_query(context, bgp_speaker_id)
@@ -568,7 +571,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_binding(self, context,
                                                 network_id, bgp_speaker_id):
         """Get all floating IP host routes for the given network binding."""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             # Query the DB for floating IP's and the IP address of the
             # gateway port
             dest_alias = aliased(l3_db.FloatingIP,
@@ -606,7 +609,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_dvr_fip_host_routes_by_binding(self, context, network_id,
                                             bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             BgpBinding = BgpSpeakerNetworkBinding
 
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
@@ -625,7 +628,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_bgp_speaker(self, context,
                                                     bgp_speaker_id):
         """Get all the floating IP host routes advertised by a BgpSpeaker."""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             dest_alias = aliased(l3_db.FloatingIP,
                                  name='destination')
             next_hop_alias = aliased(models_v2.IPAllocation,
@@ -702,7 +705,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_dvr_fip_host_routes_by_bgp_speaker(self, context,
                                                 bgp_speaker_id):
         router_attrs = l3_attrs_db.RouterExtraAttributes
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
             fip_query = self._get_fip_query(context, bgp_speaker_id)
 
@@ -731,7 +734,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                               network_id, bgp_speaker_id):
         """Get all tenant network routes for the given network."""
 
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             tenant_networks_query = self._tenant_networks_by_network_query(
                                                                context,
                                                                network_id,
@@ -750,7 +753,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                              bgp_speaker_id):
         """Get all tenant network routes with the given router as nexthop."""
 
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             scopes = self._get_address_scope_ids_for_bgp_speaker(
                                                                context,
                                                                bgp_speaker_id)
@@ -797,7 +800,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                                   bgp_speaker_id):
         """Get all tenant network routes to be advertised by a BgpSpeaker."""
 
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             tenant_nets_q = self._tenant_networks_by_bgp_speaker_query(
                                                                context,
                                                                bgp_speaker_id)
@@ -934,7 +937,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
            models_v2.IPAllocation.subnet_id == models_v2.Subnet.id]
 
     def _tenant_prefixes_by_router(self, context, router_id, bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(models_v2.Subnet.cidr.distinct())
             filters = self._tenant_prefixes_by_router_filters(router_id,
                                                               bgp_speaker_id)
@@ -962,7 +965,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                              context,
                                              router_port_id,
                                              bgp_speaker_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(models_v2.Subnet.cidr.distinct())
             filters = self._tenant_prefixes_by_router_filters(router_port_id,
                                                               bgp_speaker_id)
@@ -985,7 +988,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _bgp_speakers_for_gateway_network(self, context, network_id):
         """Return all BgpSpeakers for the given gateway network"""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(BgpSpeaker)
             query = query.filter(
                   BgpSpeakerNetworkBinding.network_id == network_id,
@@ -995,7 +998,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _bgp_speakers_for_gw_network_by_family(self, context,
                                                network_id, ip_version):
         """Return the BgpSpeaker by given gateway network and ip_version"""
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(context):
             query = context.session.query(BgpSpeaker)
             query = query.filter(
                   BgpSpeakerNetworkBinding.network_id == network_id,
@@ -1017,3 +1020,66 @@ class BgpDbMixin(common_db.CommonDbMixin):
         """Return the list of host routes given a list of (IP, nexthop)"""
         return ({'destination': x + '/32',
                  'next_hop': y} for x, y in ip_next_hop_tuples)
+
+    def _get_router(self, context, router_id):
+        try:
+            router = self._get_by_id(context, l3_db.Router, router_id)
+        except sa_exc.NoResultFound:
+            raise l3_exc.RouterNotFound(router_id=router_id)
+        return router
+
+    def _get_fip_next_hop(self, context, router_id, ip_address=None):
+        router = self._get_router(context, router_id)
+        gw_port = router.gw_port
+        if not gw_port:
+            return
+
+        if l3_dvr_db.is_distributed_router(router) and ip_address:
+            return self._get_dvr_fip_next_hop(context, ip_address)
+
+        for fixed_ip in gw_port.fixed_ips:
+            addr = netaddr.IPAddress(fixed_ip.ip_address)
+            if addr.version == 4:
+                return fixed_ip.ip_address
+
+    def _get_dvr_fip_agent_gateway_query(self, context):
+        ML2PortBinding = ml2_models.PortBinding
+        IpAllocation = models_v2.IPAllocation
+        Port = models_v2.Port
+        base_query = context.session.query(Port.network_id,
+                                           ML2PortBinding.host,
+                                           IpAllocation.ip_address)
+
+        gw_query = base_query.filter(
+            ML2PortBinding.port_id == Port.id,
+            IpAllocation.port_id == Port.id,
+            Port.device_owner == lib_consts.DEVICE_OWNER_AGENT_GW)
+        return gw_query
+
+    def _get_fip_fixed_port_host_query(self, context, fip_address):
+        ML2PortBinding = ml2_models.PortBinding
+
+        fip_query = context.session.query(
+            l3_db.FloatingIP.floating_network_id,
+            ML2PortBinding.host,
+            l3_db.FloatingIP.floating_ip_address)
+        fip_query = fip_query.filter(
+            l3_db.FloatingIP.fixed_port_id == ML2PortBinding.port_id,
+            l3_db.FloatingIP.floating_ip_address == fip_address)
+        return fip_query
+
+    def _get_dvr_fip_next_hop(self, context, fip_address):
+        try:
+            dvr_agent_gw_query = self._get_dvr_fip_agent_gateway_query(
+                context)
+            fip_fix_port_query = self._get_fip_fixed_port_host_query(
+                context, fip_address)
+            q = self._join_fip_by_host_binding_to_agent_gateway(
+                context,
+                fip_fix_port_query.subquery(),
+                dvr_agent_gw_query.subquery()).one()
+            return q[1]
+        except sa_exc.NoResultFound:
+            return
+        except sa_exc.MultipleResultsFound:
+            return
