@@ -15,6 +15,10 @@
 
 import collections
 
+from neutron_lib.agent import constants as agent_consts
+from neutron_lib.agent import topics
+from neutron_lib import context
+from neutron_lib.utils import runtime
 from oslo_config import cfg
 from oslo_log import log as logging
 import oslo_messaging
@@ -23,11 +27,8 @@ from oslo_service import periodic_task
 from oslo_utils import importutils
 
 from neutron.agent import rpc as agent_rpc
-from neutron.common import constants as n_const
 from neutron.common import rpc as n_rpc
-from neutron.common import topics
 from neutron.common import utils
-from neutron import context
 from neutron import manager
 
 from neutron_dynamic_routing.extensions import bgp as bgp_ext
@@ -43,10 +44,10 @@ class BgpDrAgent(manager.Manager):
 
     Note that the public methods of this class are exposed as the server side
     of an rpc interface.  The neutron server uses
-    neutron.api.rpc.agentnotifiers.bgp_dr_rpc_agent_api.
-    BgpDrAgentNotifyApi as the client side to execute the methods
-    here. For more information about changing rpc interfaces, see
-    http://docs.openstack.org/developer/neutron/devref/rpc_api.html.
+    api.rpc.agentnotifiers.bgp_dr_rpc_agent_api.BgpDrAgentNotifyApi as the
+    client side to execute the methods here. For more information about
+    changing rpc interfaces, see
+    https://docs.openstack.org/neutron/latest/contributor/internals/rpc_api.html.
 
     API version history:
         1.0 initial Version
@@ -94,7 +95,7 @@ class BgpDrAgent(manager.Manager):
         self.sync_state(self.context)
         self.periodic_resync(self.context)
 
-    @utils.synchronized('bgp-dragent')
+    @runtime.synchronized('bgp-dragent')
     def sync_state(self, context, full_sync=None, bgp_speakers=None):
         try:
             hosted_bgp_speakers = self.plugin_rpc.get_bgp_speakers(context)
@@ -174,7 +175,7 @@ class BgpDrAgent(manager.Manager):
         LOG.debug("Started periodic resync.")
         self._periodic_resync_helper(context)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_speaker_create_end(self, context, payload):
         """Handle bgp_speaker_create_end notification event."""
         bgp_speaker_id = payload['bgp_speaker']['id']
@@ -183,9 +184,9 @@ class BgpDrAgent(manager.Manager):
                   {'speaker_id': bgp_speaker_id})
         self.add_bgp_speaker_helper(bgp_speaker_id)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_speaker_remove_end(self, context, payload):
-        """Handle bgp_speaker_create_end notification event."""
+        """Handle bgp_speaker_remove_end notification event."""
 
         bgp_speaker_id = payload['bgp_speaker']['id']
         LOG.debug('Received BGP speaker remove notification for '
@@ -193,7 +194,7 @@ class BgpDrAgent(manager.Manager):
                   {'speaker_id': bgp_speaker_id})
         self.remove_bgp_speaker_from_dragent(bgp_speaker_id)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_peer_association_end(self, context, payload):
         """Handle bgp_peer_association_end notification event."""
 
@@ -206,7 +207,7 @@ class BgpDrAgent(manager.Manager):
                    'peer_id': bgp_peer_id})
         self.add_bgp_peer_helper(bgp_speaker_id, bgp_peer_id)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_peer_disassociation_end(self, context, payload):
         """Handle bgp_peer_disassociation_end notification event."""
 
@@ -219,7 +220,7 @@ class BgpDrAgent(manager.Manager):
                    'peer_ip': bgp_peer_ip})
         self.remove_bgp_peer_from_bgp_speaker(bgp_speaker_id, bgp_peer_ip)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_routes_advertisement_end(self, context, payload):
         """Handle bgp_routes_advertisement_end notification event."""
 
@@ -230,7 +231,7 @@ class BgpDrAgent(manager.Manager):
         routes = payload['advertise_routes']['routes']
         self.add_routes_helper(bgp_speaker_id, routes)
 
-    @utils.synchronized('bgp-dr-agent')
+    @runtime.synchronized('bgp-dr-agent')
     def bgp_routes_withdrawal_end(self, context, payload):
         """Handle bgp_routes_withdrawal_end notification event."""
 
@@ -399,10 +400,12 @@ class BgpDrAgent(manager.Manager):
 
         LOG.debug('Calling driver interface for adding BGP peer %(peer_ip)s '
                   'remote_as=%(remote_as)s to BGP Speaker running for '
-                  'local_as=%(local_as)d',
+                  'local_as=%(local_as)d '
+                  'auth_type=%(auth_type)s',
                   {'peer_ip': bgp_peer['peer_ip'],
                    'remote_as': bgp_peer['remote_as'],
-                   'local_as': bgp_speaker_as})
+                   'local_as': bgp_speaker_as,
+                   'auth_type': bgp_peer['auth_type']})
         try:
             self.dr_driver_cls.add_bgp_peer(bgp_speaker_as,
                                             bgp_peer['peer_ip'],
@@ -528,7 +531,7 @@ class BgpDrPluginApi(object):
 
     This class implements the client side of an rpc interface.
     The server side of this interface can be found in
-    neutron.api.rpc.handlers.bgp_speaker_rpc.BgpSpeakerRpcCallback.
+    api.rpc.handlers.bgp_speaker_rpc.BgpSpeakerRpcCallback.
     For more information about changing rpc interfaces, see
     doc/source/devref/rpc_api.rst.
 
@@ -627,6 +630,7 @@ class BgpSpeakerCache(object):
         routes = self.cache[bgp_speaker_id]['advertised_routes']
         for r in routes:
             if r['destination'] == route['destination'] and (
+                    route['next_hop'] is None or
                     r['next_hop'] == route['next_hop']):
                 return True
         return False
@@ -680,7 +684,7 @@ class BgpDrAgentWithStateReport(BgpDrAgent):
             ctx = context.get_admin_context_without_session()
             agent_status = self.state_rpc.report_state(ctx, self.agent_state,
                                                        True)
-            if agent_status == n_const.AGENT_REVIVED:
+            if agent_status == agent_consts.AGENT_REVIVED:
                 LOG.info(_LI("Agent has just been revived. "
                              "Scheduling full sync"))
                 self.schedule_full_resync(
