@@ -14,23 +14,25 @@
 
 from netaddr import IPAddress
 
+from neutron_lib.callbacks import events
+from neutron_lib.callbacks import registry
+from neutron_lib.callbacks import resources
 from neutron_lib import constants as n_const
+from neutron_lib import context
+from neutron_lib.services import base as service_base
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import importutils
 
-from neutron.callbacks import events
-from neutron.callbacks import registry
-from neutron.callbacks import resources
 from neutron.common import rpc as n_rpc
-from neutron import context
-from neutron.services import service_base
 
 from neutron_dynamic_routing.api.rpc.agentnotifiers import bgp_dr_rpc_agent_api  # noqa
+from neutron_dynamic_routing.api.rpc.callbacks import resources as dr_resources
 from neutron_dynamic_routing.api.rpc.handlers import bgp_speaker_rpc as bs_rpc
 from neutron_dynamic_routing.db import bgp_db
 from neutron_dynamic_routing.db import bgp_dragentscheduler_db
 from neutron_dynamic_routing.extensions import bgp as bgp_ext
+from neutron_dynamic_routing.extensions import bgp_4byte_asn
 from neutron_dynamic_routing.extensions import bgp_dragentscheduler as dras_ext
 from neutron_dynamic_routing.services.bgp.common import constants as bgp_consts
 
@@ -43,7 +45,8 @@ class BgpPlugin(service_base.ServicePluginBase,
                 bgp_dragentscheduler_db.BgpDrAgentSchedulerDbMixin):
 
     supported_extension_aliases = [bgp_ext.BGP_EXT_ALIAS,
-                                   dras_ext.BGP_DRAGENT_SCHEDULER_EXT_ALIAS]
+                                   dras_ext.BGP_DRAGENT_SCHEDULER_EXT_ALIAS,
+                                   bgp_4byte_asn.BGP_4BYTE_ASN_EXT_ALIAS]
 
     def __init__(self):
         super(BgpPlugin, self).__init__()
@@ -51,9 +54,7 @@ class BgpPlugin(service_base.ServicePluginBase,
             cfg.CONF.bgp_drscheduler_driver)
         self._setup_rpc()
         self._register_callbacks()
-
-    def get_plugin_name(self):
-        return PLUGIN_NAME
+        self.add_periodic_dragent_status_check()
 
     def get_plugin_type(self):
         return bgp_ext.BGP_EXT_ALIAS
@@ -61,11 +62,11 @@ class BgpPlugin(service_base.ServicePluginBase,
     def get_plugin_description(self):
         """returns string description of the plugin."""
         return ("BGP dynamic routing service for announcement of next-hops "
-                "for tenant networks, floating IP's, and DVR host routes.")
+                "for project networks, floating IP's, and DVR host routes.")
 
     def _setup_rpc(self):
         self.topic = bgp_consts.BGP_PLUGIN
-        self.conn = n_rpc.create_connection()
+        self.conn = n_rpc.Connection()
         self.agent_notifiers[bgp_consts.AGENT_TYPE_BGP_ROUTING] = (
             bgp_dr_rpc_agent_api.BgpDrAgentNotifyApi()
         )
@@ -115,6 +116,10 @@ class BgpPlugin(service_base.ServicePluginBase,
     def create_bgp_speaker(self, context, bgp_speaker):
         bgp_speaker = super(BgpPlugin, self).create_bgp_speaker(context,
                                                                 bgp_speaker)
+        payload = {'plugin': self, 'context': context,
+                   'bgp_speaker': bgp_speaker}
+        registry.notify(dr_resources.BGP_SPEAKER, events.AFTER_CREATE,
+                        self, payload=payload)
         return bgp_speaker
 
     def update_bgp_speaker(self, context, bgp_speaker_id, bgp_speaker):
@@ -223,19 +228,25 @@ class BgpPlugin(service_base.ServicePluginBase,
         ctx = context.get_admin_context()
         new_router_id = kwargs['router_id']
         last_router_id = kwargs['last_known_router_id']
-        next_hop = kwargs['next_hop']
-        dest = kwargs['floating_ip_address'] + '/32'
+        floating_ip_address = kwargs['floating_ip_address']
+        dest = floating_ip_address + '/32'
         bgp_speakers = self._bgp_speakers_for_gw_network_by_family(
             ctx,
             kwargs['floating_network_id'],
             n_const.IP_VERSION_4)
 
         if last_router_id and new_router_id != last_router_id:
+            # Here gives the old route next_hop a `None` value, then
+            # the DR agent side will withdraw it.
+            old_host_route = {'destination': dest, 'next_hop': None}
             for bgp_speaker in bgp_speakers:
                 self.stop_route_advertisements(ctx, self._bgp_rpc,
-                                               bgp_speaker.id, [dest])
+                                               bgp_speaker.id,
+                                               [old_host_route])
 
-        if next_hop and new_router_id != last_router_id:
+        if new_router_id and new_router_id != last_router_id:
+            next_hop = self._get_fip_next_hop(
+                ctx, new_router_id, floating_ip_address)
             new_host_route = {'destination': dest, 'next_hop': next_hop}
             for bgp_speaker in bgp_speakers:
                 self.start_route_advertisements(ctx, self._bgp_rpc,
