@@ -13,21 +13,20 @@
 # under the License.
 
 import contextlib
+
+import mock
 import netaddr
-from oslo_config import cfg
-
-from neutron_lib import constants as n_const
-from neutron_lib import exceptions as n_exc
-from oslo_utils import uuidutils
-
 from neutron.db import l3_dvr_ha_scheduler_db
-from neutron.db import l3_hamode_db
-from neutron.extensions import external_net
-from neutron.extensions import portbindings
-from neutron import manager
-from neutron.plugins.common import constants as p_const
 from neutron.tests.unit.extensions import test_l3
 from neutron.tests.unit.plugins.ml2 import test_plugin
+from neutron_lib.api.definitions import external_net
+from neutron_lib.api.definitions import portbindings
+from neutron_lib import constants as n_const
+from neutron_lib import exceptions as n_exc
+from neutron_lib.plugins import constants as plugin_constants
+from neutron_lib.plugins import directory
+from oslo_config import cfg
+from oslo_utils import uuidutils
 
 from neutron_dynamic_routing.extensions import bgp
 from neutron_dynamic_routing.services.bgp import bgp_plugin
@@ -39,7 +38,6 @@ IMAGINARY = '2b2334c8-adfe-42d9-82c6-ad866c7fc5d8'  # non existent resource id
 
 
 class TestL3Plugin(test_l3.TestL3NatAgentSchedulingServicePlugin,
-                   l3_hamode_db.L3_HA_NAT_db_mixin,
                    l3_dvr_ha_scheduler_db.L3_DVR_HA_scheduler_db_mixin):
     pass
 
@@ -102,6 +100,14 @@ class BgpEntityCreationMixin(object):
                                 router_distributed=False):
         pass
 
+    @contextlib.contextmanager
+    def gw_network(self, name='ext-net', external=False, **kwargs):
+        with self.network(name=name, **kwargs) as gw_network:
+            if external:
+                self._update('networks', gw_network['network']['id'],
+                             {'network': {external_net.EXTERNAL: True}})
+            yield gw_network
+
     @contextlib.contextmanager
     def router(self, name='bgp-test-router', tenant_id=_uuid(),
                admin_state_up=True, **kwargs):
@@ -130,7 +136,8 @@ class BgpEntityCreationMixin(object):
         if address_scope:
             subnetpool_args['address_scope_id'] = address_scope['id']
 
-        with self.network() as ext_net, self.network() as int_net,\
+        with self.gw_network(external=True) as ext_net,\
+            self.network() as int_net,\
             self.subnetpool(prefixes, **subnetpool_args) as pool:
             subnetpool_id = pool['subnetpool']['id']
             gw_net_id = ext_net['network']['id']
@@ -142,8 +149,6 @@ class BgpEntityCreationMixin(object):
                             cidr=tenant_prefix,
                             subnetpool_id=subnetpool_id,
                             ip_version=tenant_ip_net.version) as int_subnet:
-                self._update('networks', gw_net_id,
-                             {'network': {external_net.EXTERNAL: True}})
                 ext_gw_info = {'network_id': gw_net_id}
                 with self.router(external_gateway_info=ext_gw_info,
                                  distributed=distributed,
@@ -157,22 +162,7 @@ class BgpEntityCreationMixin(object):
                     yield router, ext_net, int_net
 
 
-class BgpTests(test_plugin.Ml2PluginV2TestCase,
-               BgpEntityCreationMixin):
-    fmt = 'json'
-
-    def setup_parent(self):
-        self.l3_plugin = ('neutron_dynamic_routing.tests.unit.db.test_bgp_db.'
-                          'TestL3Plugin')
-        super(BgpTests, self).setup_parent()
-
-    def setUp(self):
-        super(BgpTests, self).setUp()
-        self.l3plugin = manager.NeutronManager.get_service_plugins().get(
-            p_const.L3_ROUTER_NAT)
-        self.bgp_plugin = bgp_plugin.BgpPlugin()
-        self.plugin = manager.NeutronManager.get_plugin()
-
+class BgpTests(BgpEntityCreationMixin):
     @contextlib.contextmanager
     def subnetpool_with_address_scope(self, ip_version, prefixes=None,
                                       shared=False, admin=True,
@@ -409,7 +399,7 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
         with self.subnetpool_with_address_scope(4,
                                                 prefixes=['8.0.0.0/8']) as sp:
             with self.bgp_speaker(sp['ip_version'], 1234) as speaker,\
-                self.network() as network:
+                self.gw_network() as network:
                 network_id = network['network']['id']
                 self.bgp_plugin.add_gateway_network(self.context,
                                                    speaker['id'],
@@ -421,21 +411,18 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
 
     def test_create_bgp_speaker_with_network(self):
         with self.subnetpool_with_address_scope(4,
-                                                prefixes=['8.0.0.0/8']) as sp:
-            network = self.plugin.create_network(self.context,
-                                                {'network':
-                                                    {'name': 'test-net',
-                                                     'tenant_id': _uuid(),
-                                                     'admin_state_up': True,
-                                                     'shared': True}})
+                                                prefixes=['8.0.0.0/8']) as sp,\
+            self.gw_network(name='test-net', tenant_id=_uuid(),
+                            shared=True) as network:
+            network_id = network['network']['id']
             with self.bgp_speaker(sp['ip_version'], 1234,
-                                  networks=[network['id']]) as speaker:
+                                  networks=[network_id]) as speaker:
                 self.assertEqual(1, len(speaker['networks']))
-                self.assertTrue(network['id'] in speaker['networks'])
+                self.assertTrue(network_id in speaker['networks'])
 
     def test_remove_gateway_network(self):
-        with self.network() as network1,\
-            self.network() as network2,\
+        with self.gw_network() as network1,\
+            self.gw_network() as network2,\
             self.subnetpool_with_address_scope(4,
                                                prefixes=['8.0.0.0/8']) as sp:
             network1_id = network1['network']['id']
@@ -475,7 +462,7 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
                                                 prefixes=['8.0.0.0/8']) as sp:
             with self.bgp_speaker(sp['ip_version'], 1234) as speaker1,\
                 self.bgp_speaker(sp['ip_version'], 4321) as speaker2,\
-                self.network() as network:
+                self.gw_network() as network:
                 network_id = network['network']['id']
                 self.bgp_plugin.add_gateway_network(self.context,
                                                    speaker1['id'],
@@ -513,8 +500,8 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
             self.subnetpool_with_address_scope(4,
                                                prefixes=prefixes3,
                                                tenant_id=tenant_id) as sp3,\
-            self.network() as network1, self.network() as network2,\
-            self.network() as network3:
+            self.gw_network() as network1, self.gw_network() as network2,\
+            self.gw_network() as network3:
             network1_id = network1['network']['id']
             network2_id = network2['network']['id']
             network3_id = network3['network']['id']
@@ -725,12 +712,11 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
                                        prefixes=[binding_cidr]) as ext_pool,\
             self.subnetpool_with_address_scope(6, tenant_id=_uuid(),
                                        prefixes=[tenant_cidr]) as int_pool,\
-            self.network() as ext_net, self.network() as int_net:
+            self.gw_network(external=True) as ext_net,\
+            self.network() as int_net:
             gw_net_id = ext_net['network']['id']
             ext_pool_id = ext_pool['id']
             int_pool_id = int_pool['id']
-            self._update('networks', gw_net_id,
-                         {'network': {external_net.EXTERNAL: True}})
             with self.subnet(cidr=None,
                              subnetpool_id=ext_pool_id,
                              network=ext_net,
@@ -1097,11 +1083,10 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
                 self.assertTrue(fip_prefix_found)
 
     def test__bgp_speakers_for_gateway_network_by_ip_version(self):
-        with self.network() as ext_net, self.bgp_speaker(6, 1234) as s1,\
+        with self.gw_network(external=True) as ext_net,\
+            self.bgp_speaker(6, 1234) as s1,\
             self.bgp_speaker(6, 4321) as s2:
             gw_net_id = ext_net['network']['id']
-            self._update('networks', gw_net_id,
-                         {'network': {external_net.EXTERNAL: True}})
             self.bgp_plugin.add_gateway_network(self.context,
                                                 s1['id'],
                                                 {'network_id': gw_net_id})
@@ -1115,11 +1100,10 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
             self.assertEqual(2, len(speakers))
 
     def test__bgp_speakers_for_gateway_network_by_ip_version_no_binding(self):
-        with self.network() as ext_net, self.bgp_speaker(6, 1234),\
+        with self.gw_network(external=True) as ext_net,\
+            self.bgp_speaker(6, 1234),\
             self.bgp_speaker(6, 4321):
             gw_net_id = ext_net['network']['id']
-            self._update('networks', gw_net_id,
-                         {'network': {external_net.EXTERNAL: True}})
             speakers = self.bgp_plugin._bgp_speakers_for_gw_network_by_family(
                                                                  self.context,
                                                                  gw_net_id,
@@ -1155,17 +1139,74 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
 
     def _create_scenario_test_fips(self, ext_net_id,
                                    tenant_id, port_ids):
+        fips = []
         for port_id in port_ids:
             fip_data = {'floatingip': {'floating_network_id': ext_net_id,
                                        'tenant_id': tenant_id,
                                        'port_id': port_id}}
-            self.l3plugin.create_floatingip(self.context, fip_data)
+            fip = self.l3plugin.create_floatingip(self.context, fip_data)
+            fips.append(fip)
+        return fips
+
+    def test_floatingip_update_callback(self):
+        gw_prefix = '172.16.10.0/24'
+        tenant_prefix = '10.10.10.0/24'
+        tenant_id = _uuid()
+        agent_confs = [{"host": "network1", "mode": "legacy"}]
+        self._create_scenario_test_l3_agents(agent_confs)
+        with self.router_with_external_and_tenant_networks(
+                tenant_id=tenant_id,
+                gw_prefix=gw_prefix,
+                tenant_prefix=tenant_prefix) as res:
+            router, ext_net, int_net = res
+            gw_net_id = ext_net['network']['id']
+            ext_gw_info = router['external_gateway_info']
+            next_hop = ext_gw_info[
+                'external_fixed_ips'][0]['ip_address']
+
+            port_configs = [{'net_id': int_net['network']['id'],
+                             'host': 'compute1'}]
+            ports = self._create_scenario_test_ports(tenant_id, port_configs)
+            port_ids = [port['id'] for port in ports]
+            with self.bgp_speaker(4, 1234, networks=[gw_net_id]) as speaker:
+                with mock.patch.object(
+                    self.bgp_plugin,
+                    'start_route_advertisements') as start_adv:
+                    with mock.patch.object(
+                        self.bgp_plugin,
+                        'stop_route_advertisements') as stop_adv:
+                        bgp_speaker_id = speaker['id']
+                        # Create and associate floating IP
+                        fips = self._create_scenario_test_fips(
+                            gw_net_id, tenant_id, port_ids)
+                        fip_id = fips[0]['id']
+                        fip_addr = fips[0]['floating_ip_address']
+                        host_route = {'destination': fip_addr + '/32',
+                                      'next_hop': next_hop}
+                        start_adv.assert_called_once_with(
+                            mock.ANY,
+                            mock.ANY,
+                            bgp_speaker_id,
+                            [host_route])
+                        fip_data = {'floatingip': {'port_id': None}}
+                        fip = self.l3plugin.update_floatingip(
+                            self.context,
+                            fip_id,
+                            fip_data)
+                        self.assertIsNone(fip['port_id'])
+                        # Dissociate floating IP,
+                        # withdraw route with next_hop None
+                        host_route['next_hop'] = None
+                        stop_adv.assert_called_once_with(
+                            mock.ANY,
+                            mock.ANY,
+                            bgp_speaker_id,
+                            [host_route])
 
     def _test_legacy_router_fips_next_hop(self, router_ha=False):
         if router_ha:
             cfg.CONF.set_override('l3_ha', True)
             cfg.CONF.set_override('max_l3_agents_per_router', 2)
-            cfg.CONF.set_override('min_l3_agents_per_router', 2)
         gw_prefix = '172.16.10.0/24'
         tenant_prefix = '10.10.10.0/24'
         tenant_id = _uuid()
@@ -1215,3 +1256,64 @@ class BgpTests(test_plugin.Ml2PluginV2TestCase,
 
     def test_ha_router_fips_has_no_next_hop_to_fip_agent_gateway(self):
         self._test_legacy_router_fips_next_hop(router_ha=True)
+
+    def _test__get_fip_next_hop(self, distributed=False):
+        gw_prefix = '172.16.10.0/24'
+        tenant_prefix = '10.10.10.0/24'
+        tenant_id = _uuid()
+        agent_confs = [{"host": "compute1", "mode": "dvr"},
+                       {"host": "network1", "mode": "dvr_snat"}]
+        self._create_scenario_test_l3_agents(agent_confs)
+        with self.router_with_external_and_tenant_networks(
+                tenant_id=tenant_id,
+                gw_prefix=gw_prefix,
+                tenant_prefix=tenant_prefix,
+                distributed=distributed) as res:
+            router, ext_net, int_net = res
+            ext_gw_info = router['external_gateway_info']
+            legacy_gw_ip = ext_gw_info[
+                'external_fixed_ips'][0]['ip_address']
+            gw_net_id = ext_net['network']['id']
+            # Whatever the router type is, we always create such dvr fip agent
+            # gateway port, in order to make sure that legacy router fip bgp
+            # route next_hop is not the dvr_fip_agent_gw.
+            fip_gw = self.l3plugin.create_fip_agent_gw_port_if_not_exists(
+                self.context, gw_net_id, 'compute1')
+            port_configs = [{'net_id': int_net['network']['id'],
+                             'host': 'compute1'}]
+            ports = self._create_scenario_test_ports(tenant_id, port_configs)
+            dvr_gw_ip = fip_gw['fixed_ips'][0]['ip_address']
+            fip_data = {'floatingip': {'floating_network_id': gw_net_id,
+                                       'tenant_id': tenant_id,
+                                       'port_id': ports[0]['id']}}
+            fip = self.l3plugin.create_floatingip(self.context, fip_data)
+            fip_address = fip['floating_ip_address']
+            with self.bgp_speaker(4, 1234, networks=[gw_net_id]):
+                next_hop = self.bgp_plugin._get_fip_next_hop(
+                    self.context, router['id'], fip_address)
+                if distributed:
+                    self.assertEqual(dvr_gw_ip, next_hop)
+                else:
+                    self.assertEqual(legacy_gw_ip, next_hop)
+
+    def test__get_fip_next_hop_legacy(self):
+        self._test__get_fip_next_hop()
+
+    def test__get_fip_next_hop_dvr(self):
+        self._test__get_fip_next_hop(distributed=True)
+
+
+class Ml2BgpTests(test_plugin.Ml2PluginV2TestCase,
+                  BgpTests):
+    fmt = 'json'
+
+    def setup_parent(self):
+        self.l3_plugin = ('neutron_dynamic_routing.tests.unit.db.test_bgp_db.'
+                          'TestL3Plugin')
+        super(Ml2BgpTests, self).setup_parent()
+
+    def setUp(self):
+        super(Ml2BgpTests, self).setUp()
+        self.l3plugin = directory.get_plugin(plugin_constants.L3)
+        self.bgp_plugin = bgp_plugin.BgpPlugin()
+        self.plugin = directory.get_plugin()
